// Parser.cs
using System;
using System.Collections;


namespace TtLisp {
  // ---------------------------------------------------------------------
  // -- class Parser -----------------------------------------------------
  public class Parser {
    // -- static ---------------------------------------------------------
    public static SymbolicExpression[] Parse( string src ) {
      return new Parser( src ).Parse();
    }

    // -- instance -------------------------------------------------------
    private Scanner scanner;

    private Parser( string src ) {
      this.scanner = new Scanner( src );
    }

    private SymbolicExpression[] Parse() {
      ArrayList ary = new ArrayList();
      while ( this.scanner.MoveNext() ) {
        ary.Add( this.ParseOne() );
      }
      SymbolicExpression[] _tmp = new SymbolicExpression[ary.Count];
      ary.CopyTo( _tmp );
      return _tmp;
    }

    private SymbolicExpression ParseOne() {
      switch ( this.scanner.Current ) {
      case "(":
        return this.ParseAsList();

      case ")":
        throw new InvalidEncloseListException();

      case ".":
        throw new CanNotParseCharacterException();

      case "'":
        return this.ParseAsQuote();

      case "#'":
        return this.ParseAsFunction();

      default:
        if ( this.scanner.Current[0] == '"' ) {
          return this.ParseAsLString();
        }
        return this.ParseAsAtom();
      }
    }

    private SymbolicExpression ParseAsList() {
      if ( !(this.scanner.MoveNext()) ) {
        throw new UnexpectedEOFException();
      }
      if ( this.scanner.Current == ")" ) {
        return Nil.NIL;
      }
      else if ( this.scanner.Current == "." ) {
        if ( !(this.scanner.MoveNext()) ) {
          throw new UnexpectedEOFException();
        }
        SymbolicExpression _tmp = this.ParseOne();
        if ( !(this.scanner.MoveNext()) ) {
          throw new UnexpectedEOFException();
        }
        else if ( this.scanner.Current != ")" ) {
          throw new InvalidEncloseListException();
        }
        return _tmp;
      }
      else {
        return new List( this.ParseOne(), this.ParseAsList() );
      }
    }

    private List ParseAsQuote() {
      if ( !(this.scanner.MoveNext()) ) {
        throw new UnexpectedEOFException();
      }
      return List.Quote( this.ParseOne() );
    }

    private List ParseAsFunction() {
      if ( !(this.scanner.MoveNext()) ) {
        throw new UnexpectedEOFException();
      }
      return List.Make( new Symbol( "function" ), this.ParseOne() );
    }

    private LString ParseAsLString() {
      return new LString( this.scanner.Current.Substring( 1, this.scanner.Current.Length - 2 ) );
    }

    private SymbolicExpression ParseAsAtom() {
      try {
        return new Integer( Convert.ToInt32( this.scanner.Current ) );
      }
      catch ( FormatException ) {
        return this.ParseAsSymbol();
      }
      catch ( OverflowException ) {
        return this.ParseAsSymbol();
      }
    }

    private SymbolicExpression ParseAsSymbol() {
      switch( this.scanner.Current ) {
      case "nil":
        return Nil.NIL;
      case "t":
        return True.T;
      default:
        return new Symbol( this.scanner.Current );
      }
    }
  }

  // ---------------------------------------------------------------------
  // -- class Scanner ----------------------------------------------------
  public class Scanner {
    // -- instance -------------------------------------------------------
    private string src;
    private string current;
    private int pos;

    public Scanner( string src ) {
      this.src = src;
      this.pos = 0;
      this.current = "";
    }

    public string GetNext() { this.MoveNext(); return this.Current; }

    public string Current { get { return this.current; } }
    public bool IsFinished { get { return this.current == null; } }
    public bool MoveNext() {
      this.current = null;

      int start = this.pos;
      // 󔒂ǂݔ΂
      for (;;) {
        if ( start >= this.src.Length ) { return false; }
        else if ( Char.IsWhiteSpace( this.src, start ) ) { start += 1; }
        else if ( this.src[start] == ';' ) {
          for (;;) {
            if ( start >= this.src.Length ) { return false; }
            if ( this.src[start] == '\n' ) { break; }
            start += 1;
          }
        }
        else { break; }
      }

      int length;
      switch ( this.src[start] ) {
      case '|':
      case ',':
      case '\\':
      case ':':
      case '`':
        // ݂͕@ƂĖ
        throw new CanNotParseCharacterException();

      case '(':
      case ')':
      case '\'':
        length = 1;
        break;

      case '#':
        length = this.SearchLengthAsSharp( start );
        break;
        
      case '"':
        length = this.SearchLengthAsLString( start );
        break;

      default:
        length = this.SearchLengthAsAtom( start );
        break;
      }

      this.current = this.src.Substring( start, length );
      this.pos = start + length;
      return true;
    }

    private int SearchLengthAsSharp( int start ) {
      switch( this.src[start + 1] ) {
      case '\'':
        return 2;

      default:
        throw new CanNotParseCharacterException();
      }
    }

    private int SearchLengthAsLString( int start ) {
      for ( int length = 1; start + length <= this.src.Length; length += 1 ) {
        if ( this.src[start + length] == '"' ) {
          return length + 1;
        }
      }
      throw new NotEncloseLStringException();
    }

    private int SearchLengthAsAtom( int start ) {
      for ( int length = 1; ; length += 1 ) {
        if ( start + length >= this.src.Length ) {
          return length;
        }
        if ( Char.IsWhiteSpace( this.src[start + length] ) ) {
          return length;
        }
        switch ( this.src[start + length] ) {
        case ';':
        case '(':
        case ')':
        case '\'':
        case '|':
        case ',':
        case '#':
        case '\\':
        case ':':
        case '`':
        case '"':
          return length;
        }
      }
    }

  }
}
