// SystemFunctions.cs
using System;
using System.Collections;
using System.IO;


namespace TtLisp.SystemFunctions {

  public class FunctionCar : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size != 1 ) {
        throw new InvalidArgumentCountException();
      }
      try { return ((List)args[0]).Car; }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionCdr : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size != 1 ) {
        throw new InvalidArgumentCountException();
      }
      try { return ((List)args[0]).Cdr; }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionQuote : SpecialForm {
    public override SymbolicExpression Call( Env env, List args ) {
      if ( args.Size != 1 ) {
        throw new InvalidArgumentCountException();
      }
      return args[0];
    }
  }

  public class FunctionCons : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size != 2 ) {
        throw new InvalidArgumentCountException();
      }
      return new List( args[0], args[1] );
    }
  }

  public class FunctionAtom : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size != 1 ) {
        throw new InvalidArgumentCountException();
      }
      if ( args[0] is Nil ) { return True.T; }
      return args[0] is Atom ? True.T : SymbolicExpression.NIL;
    }
  }

  public class FunctionEq : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size != 2 ) {
        throw new InvalidArgumentCountException();
      }
      return args[0].LispEq( args[1] ) ? True.T : SymbolicExpression.NIL;
    }
  }

  public class FunctionProgn : SpecialForm {
    public override SymbolicExpression Call( Env env, List args ) {
      SymbolicExpression last = Nil.NIL;
      foreach ( SymbolicExpression s in args ) {
        last = env.Eval( s );
      }
      return last;
    }
  }

  public class FunctionIf : SpecialForm {
    public override SymbolicExpression Call( Env env, List args ) {
      if ( args.Size != 2 && args.Size != 3 ) {
        throw new InvalidArgumentCountException();
      }

      if ( env.Eval( args.Car ) is Nil ) {
        return env.Eval( args[2] );
      }
      else {
        return env.Eval( args[1] );
      }
    }
  }

  public class FunctionCond : SpecialForm {
    Function progn = new FunctionProgn();
    public override SymbolicExpression Call( Env env, List args ) {
      if ( args.Size < 1  ) {
        throw new InvalidArgumentCountException();
      }
      try {
        foreach ( List cur in args ) {
          if ( env.Eval( cur.Car ) is Nil ) {
            continue;
          }
          return this.progn.Call( env, cur.Cdr is List ? (List)cur.Cdr : Nil.NIL );
        }
        return Nil.NIL;
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionLambda : SpecialForm {
    public override SymbolicExpression Call( Env env, List args ) {
      if ( args.Size != 2 ) {
        throw new InvalidArgumentCountException();
      }
      try { return new Lambda( (List)args[0], args[1] ); }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionMacro : SpecialForm {
    public override SymbolicExpression Call( Env env, List args ) {
      if ( args.Size != 2 ) {
        throw new InvalidArgumentCountException();
      }
      try { return new Macro( (List)args[0], args[1] ); }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionFunction : SpecialForm {
    public override SymbolicExpression Call( Env env, List args ) {
      if ( args.Size != 1 ) {
        throw new InvalidArgumentCountException();
      }
      try { return env.FunctionTable[(ISymbol)args[0]]; }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionBindFunction : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size != 2 ) {
        throw new InvalidArgumentCountException();
      }
      try {
        Symbol funcsym = (Symbol)args[0];
        Function lambda = (Function)args[1];
        env.FunctionTable.Define( funcsym, lambda );
        return funcsym;
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionSet : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size != 2 ) {
        throw new InvalidArgumentCountException();
      }
      try {
        ISymbol sym = (ISymbol)args[0];
        SymbolicExpression val = args[1];
        env.SymbolTable.Intern( sym, val );
        return val;
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionApply : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size != 2 ) {
        throw new InvalidArgumentCountException();
      }
      try {
        ISymbol funcsym = (ISymbol)args[0];
        List _tmp = (List)args[1];
        List arguments = Nil.NIL;
        for ( int i = _tmp.Size - 1; i >= 0; --i ) {
          arguments = new List( List.Quote( _tmp[i] ), arguments );
        }
        return env.FunctionTable[funcsym].Call( env, arguments );
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionFuncall : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size < 1 ) {
        throw new InvalidArgumentCountException();
      }
      try {
        ISymbol funcsym = (ISymbol)args[0];
        List _tmp = (List)args.Cdr;
        List arguments = Nil.NIL;
        for ( int i = _tmp.Size - 1; i >= 0; --i ) {
          arguments = new List( List.Make( new Symbol( "quote" ), _tmp[i] ), arguments );
        }
        return env.FunctionTable[funcsym].Call( env, arguments );
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionSetQ : SpecialForm {
    public override SymbolicExpression Call( Env env, List args ) {
      if ( args.Size != 2 ) {
        throw new InvalidArgumentCountException();
      }
      try {
        ISymbol sym = (ISymbol)args[0];
        SymbolicExpression val = env.Eval( args[1] );
        env.SymbolTable.Intern( sym, val );
        return val;
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionEval : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size != 1 ) {
        throw new InvalidArgumentCountException();
      }
      env.PushStack();
      try { return env.Eval( args[0] ); }
      finally {
        env.PopStack();
      }
    }
  }

  public class FunctionList : Lambda {
    public override SymbolicExpression Call( Env env, List args ) {
      return this.EvalEachOfList( env, args );
    }
  }

  public class FunctionMacroExpand : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size != 1 ) {
        throw new InvalidArgumentCountException();
      }
      try {
        Macro macro = (Macro)env.FunctionTable[(ISymbol)((List)args[0])[0]];
        return macro.Expand( env, (List)((List)args[0]).Cdr );
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionLetAste : SpecialForm {
    Function progn = new FunctionProgn();
    public override SymbolicExpression Call( Env env, List args ) {
      SymbolTable prev = env.SymbolTable;
      env.PushStack();
      foreach ( string name in prev.Keys ) {
        ISymbol key = new Symbol( name );
        if ( !(prev.DefinedConst( key )) ) {
          env.SymbolTable.LocalIntern( key, prev[key] );
        }
      }

      try {
        // args[0] => (X (Y 100) (Z 200))
        foreach ( SymbolicExpression s in (List)args[0] ) {
          if ( s is ISymbol ) {
            env.SymbolTable.LocalIntern( (ISymbol)s, Nil.NIL );
          }
          else if ( s is List ) {
            env.SymbolTable.LocalIntern( (ISymbol)((List)s)[0], env.Eval( ((List)s)[1] ) );
          }
          else {
            throw new InvalidArgumentTypeException();
          }
        }
        return this.progn.Call( env, args.Cdr is List ? (List)args.Cdr : Nil.NIL );
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
      finally {
        env.PopStack();
      }
    }
  }

  public class FunctionIntAdd : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      try {
        Integer sum = new Integer( 0 );
        foreach ( Integer i in args ) {
          sum = sum.Add( i );
        }
        return sum;
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionIntSub : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      try {
        if ( args.Size == 1 ) {
          return new Integer( 0 ).Sub( (Integer)args[0] );
        }
        Integer sum = (Integer)args[0];
        foreach ( Integer i in (List)args.Cdr ) {
          sum = sum.Sub( i );
        }
        return sum;
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionIntMul : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      try {
        Integer sum = new Integer( 0 );
        foreach ( Integer i in args ) {
          sum = sum.Mul( i );
        }
        return sum;
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionIntDiv : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      try {
        Integer sum = args.Size == 1 ? new Integer( 1 ) : (Integer)args[0];
        List dividers = args.Size == 1 ? args : (List)args.Cdr;
        foreach ( Integer i in dividers ) {
          sum = sum.Div( i );
        }
        return sum;
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionIntMod : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      try {
        if ( args.Size != 2 ) {
          throw new InvalidArgumentCountException();
        }
        return ((Integer)args[0]).Mod( (Integer)args[1] );
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionIntEq : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      try {
        if ( args.Size < 1 ) {
          throw new InvalidArgumentCountException();
        }
        Integer master = (Integer)args[0];
        foreach ( Integer i in args ) {
          if ( master.Ne( i ) is True ) {
            return Nil.NIL;
          }
        }
        return True.T;
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }
  public class FunctionIntNe : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      try {
        if ( args.Size < 1 ) {
          throw new InvalidArgumentCountException();
        }
        Hashtable table = new Hashtable();
        foreach ( Integer i in args ) {
          if ( table.ContainsKey( i.Value ) ) {
            return Nil.NIL;
          }
          table.Add( i.Value, true );
        }
        return True.T;
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionIntGt : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      try {
        if ( args.Size < 1 ) {
          throw new InvalidArgumentCountException();
        }
        Integer prev = (Integer)args[0];
        foreach ( Integer i in (List)args.Cdr ) {
          if ( prev.Gt( i ) is Nil ) {
            return Nil.NIL;
          }
          prev = i;
        }
        return True.T;
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionIntLt : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      try {
        if ( args.Size < 1 ) {
          throw new InvalidArgumentCountException();
        }
        Integer prev = (Integer)args[0];
        foreach ( Integer i in (List)args.Cdr ) {
          if ( prev.Lt( i ) is Nil ) {
            return Nil.NIL;
          }
          prev = i;
        }
        return True.T;
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionIntGe : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      try {
        if ( args.Size < 1 ) {
          throw new InvalidArgumentCountException();
        }
        Integer prev = (Integer)args[0];
        foreach ( Integer i in (List)args.Cdr ) {
          if ( prev.Ge( i ) is Nil ) {
            return Nil.NIL;
          }
          prev = i;
        }
        return True.T;
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }
  public class FunctionIntLe : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      try {
        if ( args.Size < 1 ) {
          throw new InvalidArgumentCountException();
        }
        Integer prev = (Integer)args[0];
        foreach ( Integer i in (List)args.Cdr ) {
          if ( prev.Le( i ) is Nil ) {
            return Nil.NIL;
          }
          prev = i;
        }
        return True.T;
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionPrint : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size != 1 ) {
        throw new InvalidArgumentCountException();
      }
      Console.WriteLine( args[0] );
      return args[0];
    }
  }

  public class FunctionRequire : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size != 1 ) {
        throw new InvalidArgumentCountException();
      }
      if ( !(args[0] is LString) ) {
        throw new InvalidArgumentTypeException();
      }

      string src;
      try {
        using ( StreamReader reader = new StreamReader( ((LString)args[0]).Value ) ) {
          src = reader.ReadToEnd();
        }
      }
      catch { throw new FileNotReadException(); }

      foreach ( SymbolicExpression s in Parser.Parse( src ) ) {
        env.Eval( s );
      }
      return True.T;
    }
  }

  public class FunctionAllFunctions : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size != 0 ) {
        throw new InvalidArgumentCountException();
      }

      List list = Nil.NIL;
      foreach ( string str in env.FunctionTable.Keys ) {
        list = new List( new Symbol( str ), list );
      }
      return list;
    }
  }

  public class FunctionEvery : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size != 2 ) {
        throw new InvalidArgumentCountException();
      }

      try {
        Function func = (Function)args[0];
        foreach ( SymbolicExpression s in (List)args[1] ) {
          SymbolicExpression ret = func.Call( env, List.Make( List.Quote( s ) ) );
          if ( ret is Nil ) { return Nil.NIL; }
        }
        return True.T;
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionMapCar : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      try {
        if ( args.Size < 2 ) {
          throw new InvalidArgumentCountException();
        }

        Function func = (Function)args[0];
        ArrayList returns = new ArrayList();

        int min = Int32.MaxValue;
        foreach ( List list in (List)args.Cdr ) {
          if ( list.Size < min ) { min = list.Size; }
        }
        for ( int k = 0; k < min; ++k ) {
          List arguments = Nil.NIL;
          for ( int i = args.Size -1; i > 0; --i ) {
            arguments = new List( List.Quote( ((List)args[i])[k] ), arguments );
          }
          returns.Add( func.Call( env, arguments ) );
        }
        return List.Make( (SymbolicExpression[])returns.ToArray( typeof( SymbolicExpression ) ) );
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionAnd : Macro {
    protected override SymbolicExpression Yield( Env env, List args ) {
      try {
        if ( args is Nil ) { return True.T; }
        if ( args.Size < 2  ) { return args[0]; }
        return List.Make( new Symbol( "if" ), args[0], this.Yield( env, (List)args.Cdr ) );
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionOr : Macro {
    protected override SymbolicExpression Yield( Env env, List args ) {
      try {
        return List.Make( new Symbol( "let*" ), List.MakeAsSymbol( "_tmp" ), this.MakeOrList( env, args ) );
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
    private SymbolicExpression MakeOrList( Env env, List args ) {
      if ( args is Nil ) { return Nil.NIL; }
      if ( args.Size < 2  ) { return args[0]; }
      return List.Make( new Symbol( "if" ),
                        List.Make( new Symbol( "setq" ), new Symbol( "_tmp" ), args[0] ),
                        new Symbol( "_tmp" ),
                        this.MakeOrList( env, (List)args.Cdr ) );
    }
  }

  public class FunctionNot : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size != 1 ) {
        throw new InvalidArgumentCountException();
      }
      return args[0] is Nil ? True.T : SymbolicExpression.NIL;
    }
  }

  public class FunctionSymbolFunction : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size != 1 ) {
        throw new InvalidArgumentCountException();
      }
      return env.FunctionTable[(ISymbol)args[0]];
    }
  }

}
